/*
Copyright 2020 Sven Krauss

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
 * Some example code to show readout of Ista sensonic II using the opto interface.
 * 
 * This example uses the libmbus library. It must be installed pior.
 * 
 * Installation intructions:
 * 
 * git clone https://github.com/rscada/libmbus.git
 * cd libmbus
 * ./build.sh
 * sudo make install
 */

#include <mbus/mbus.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>

#include <time.h>
#include <errno.h>

/*
 * Send the wakeup sequence to the device.
 * 
 * The wake up sequence may differ to other devices. For ista sensonic
 * devices 0x55 must be send for 0.5s with 300Bit/s 8N1
 * 
 * @handle Pointer to the mbus_handle
 * @return 0 on success, not 0 on any error.
 */
int wakeup_ista(mbus_handle *handle){

    int ret = 0;

    struct termios tty;
    ret = tcgetattr(handle->fd, &tty);
    if (ret != 0)
        return ret;

    // Set to 300 Baud no parity
    tty.c_cflag &= ~(PARENB | PARODD);
    cfsetospeed(&tty, B300);
    cfsetispeed(&tty, B300);
    ret = tcsetattr(handle->fd, TCSANOW, &tty);
    if (ret != 0)
        return ret;

    // Send the wake up sequence
    unsigned char wakeup[16];
    memset(wakeup, 0x55, sizeof(wakeup));
    ssize_t written = write(handle->fd, wakeup, sizeof(wakeup));
    // Send out the all bytes
    ret = tcdrain(handle->fd);
    if (ret != 0)
        return ret;

    if (written != sizeof(wakeup))
        return -1;

    // Wait for 120ms
    struct timespec sleeptime =
        { .tv_sec = 0, .tv_nsec = 120 * 1000000 };

    ret = nanosleep(&sleeptime, NULL);
    if (ret != 0)
        return ret;

    // Enable parity again
    tty.c_cflag |= PARENB;
    ret = tcsetattr(handle->fd, TCSANOW, &tty);
    if (ret != 0)
        return ret;

    // Set the final baud rate
    return mbus_serial_set_baudrate(handle, 2400);
}

/*
 * Default main
 * 
 */
int main(int argc, char **argv) {

    int ret = 1;
    mbus_frame reply;
    mbus_frame_data reply_data;
    mbus_handle *handle = NULL;

    const char *interface = "/dev/ttyUSB0";
    int addr = 0xFE;

    if ((handle = mbus_context_serial(interface)) == NULL){
        fprintf(stderr, "Could not initialize M-Bus context: %s\n",  mbus_error_str());
        return 1;
    }

    mbus_register_send_event(handle, &mbus_dump_send_event);
    mbus_register_recv_event(handle, &mbus_dump_recv_event);

    if (mbus_connect(handle) == -1){
        fprintf(stderr,"Failed to setup connection to M-bus gateway\n");
        mbus_context_free(handle);
        return 1;
    }

    if (wakeup_ista(handle) != 0){
        fprintf(stderr, "Failed to wakup the device\n");
        goto out;
    }

    if (mbus_send_request_frame(handle, addr) == -1) {
        fprintf(stderr, "Failed to send M-Bus request frame.\n");
        goto out;
    }

    if (mbus_recv_frame(handle, &reply) != MBUS_RECV_RESULT_OK){
        fprintf(stderr, "Failed to receive M-Bus response frame.\n");
        goto out;
    }

    if (mbus_frame_data_parse(&reply, &reply_data) == -1){
        fprintf(stderr, "M-bus data parse error: %s\n", mbus_error_str());
        goto out;
    }

    { // Optional xml print
        char *xml_result;
        if ((xml_result = mbus_frame_data_xml(&reply_data)) == NULL){
            fprintf(stderr, "Failed to generate XML representation of MBUS frame: %s\n", mbus_error_str());
            goto out;
        }

        printf("%s", xml_result);
        free(xml_result);
    }

    // manual free
    if (reply_data.data_var.record){
        mbus_data_record_free(reply_data.data_var.record); // free's up the whole list
    }

out:
    mbus_disconnect(handle);
    mbus_context_free(handle);
    return ret;
}

